@@ -2,13 +2,13 @@ from __future__ import annotations
 
 from abc import ABC, abstractmethod
 import math
-from typing import Any, Dict, Iterable, Optional, Tuple, Union
-from warnings import warn
+from typing import Any, Dict, Iterable, Optional, Sequence, Tuple, Union
 
 import numpy as np
 import pinocchio
 
-from timor.task import Solution, Tolerance
+from timor import RobotBase
+from timor.task import Solution, Task, Tolerance
 from timor.utilities import logging
 from timor.utilities.jsonable import JSONable_mixin
 from timor.utilities.tolerated_pose import ToleratedPose
@@ -21,7 +21,7 @@ class ConstraintBase(ABC, JSONable_mixin):
     global_only: bool = False  # If this is True, the constraint cannot be used as a local goal constraint.
 
     @classmethod
-    def from_json_data(cls, d: Dict[str, Any]) -> ConstraintBase:
+    def from_json_data(cls, d: Dict[str, Any], *args, **kwargs) -> ConstraintBase:
         """
         Converts a description (string, parsed from json) to a constraint
 
@@ -50,17 +50,17 @@ class ConstraintBase(ABC, JSONable_mixin):
         # Shortcut to _from_json_data if used on subclasses to return Error or expected type if possible
         if cls is not ConstraintBase:
             if cls is type2class[constraint_type]:
-                return cls._from_json_data(d)
+                return cls._from_json_data(d, *args, **kwargs)
             raise ValueError(f"Want to create {type(cls)} with description of type {constraint_type}.")
         try:
             class_ref = type2class[constraint_type]
         except KeyError as e:
-            raise NotImplementedError(f"Unknown constraint of type {constraint_type}!") from e
+            raise AttributeError(f"Unknown constraint of type {constraint_type}!") from e
 
-        return class_ref._from_json_data(d)
+        return class_ref._from_json_data(d, *args, **kwargs)
 
     @classmethod
-    def _from_json_data(cls, description: Dict[str, any]) -> ConstraintBase:
+    def _from_json_data(cls, description: Dict[str, any], *args, **kwargs) -> ConstraintBase:
         """
         Default constraints can just be constructed from using json fields as constructor inputs.
 
@@ -68,21 +68,28 @@ class ConstraintBase(ABC, JSONable_mixin):
         """
         return cls(**description)
 
-    @abstractmethod
     def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
-        """Should be implemented to check a constraint on a subset of a solution until time t."""
-        pass
+        """Checks that a constraint is valid on a subset of a solution until time t."""
+        return all(self.is_valid_at(solution, t_i) for t_i in solution.time_steps[:solution.get_time_id(t) + 1])
+
+    def fulfilled(self, solution: Solution.SolutionBase) -> bool:
+        """Returns True iff a constraint is held over the whole solution"""
+        return self.is_valid_until(solution, solution.time_steps[-1])
 
     @abstractmethod
-    def fulfilled(self, solution: Solution.SolutionBase) -> bool:
-        """Main access point to check constraints"""
-        pass
+    def is_valid_at(self, solution: Solution.SolutionBase, t: float) -> bool:
+        """Checks that a constraint is not hurt at a specific time t"""
 
     @abstractmethod
     def to_json_data(self) -> Dict[str, any]:
         """Should be implemented to convert the constraint to a json-serializable dictionary"""
 
-    def visualize(self, viz: pinocchio.visualize.MeshcatVisualizer, scale: float = 1.) -> None:
+    @property
+    @abstractmethod
+    def _equality_parameters(self) -> Tuple[any, ...]:
+        """Returns the parameters that should be used to check equality of constraints"""
+
+    def visualize(self, viz: pinocchio.visualize.MeshcatVisualizer, scale: float = 1., show_name: bool = False) -> None:
         """Visualize the constraint in an existing meshcat window.
 
         If possible a constraint should have a visualization. The default is no visualization and should be overwritten
@@ -91,15 +98,75 @@ class ConstraintBase(ABC, JSONable_mixin):
         :param viz: A meshcat visualizer as defined in pinocchio
         :param scale: The visualization method assumes the robot size is ~1-2m maximum reach. Use the scale factor
             to adapt for significantly different sizes of robots and/or obstacles.
+        :param show_name: If True, the name of the constraint is shown in the visualization
         """
         pass
 
+    def __eq__(self, other):
+        """Checks if two constraints are equal"""
+        if type(self) is not type(other):
+            return NotImplemented  # pragma: no cover
+        return self._equality_parameters == other._equality_parameters
 
-class AlwaysTrueConstraint(ConstraintBase):
+
+class RobotConstraint(ABC):
+    """
+    This is a mixin for constraints that can be verified given a task, a robot, and its joint positions and velocities.
+
+    They are not necessarily to be used only in tasks, but can be used to check whether a robot is in a valid state for
+    downstream tasks like the computation of the inverse kinematics.
+    """
+
+    def check_single_state(self,
+                           task: Optional[Task.Task],
+                           robot: RobotBase,
+                           q: Sequence[float],
+                           dq: Sequence[float],
+                           ddq: Optional[Sequence[float]] = None) -> bool:
+        """
+        Checks whether the constraint is fulfilled for a single state of the robot.
+
+        :param task: The task that is being solved -- if not provided, an empty task will be generated.
+        :param robot: The robot that is used to solve the task.
+        :param q: The joint positions of the robot.
+        :param dq: The joint velocities of the robot.
+        :param ddq: The joint accelerations of the robot.
+        """
+
+        def reshape(arr: Sequence[float]) -> np.ndarray:
+            """Helper to reshape arrays to (1, dof)"""
+            ret = np.reshape(arr, (1, -1))
+            if not ret.shape[1] == robot.dof:
+                raise ValueError(f"Invalid shape of input argument q and/or dq: {np.asarray(arr).shape}")
+            return ret
+
+        q, dq = map(reshape, (q, dq))
+        ddq = reshape(ddq) if ddq is not None else None
+        task = task if task is not None else Task.Task.empty()
+        return self._check_single_state(task, robot, q, dq, ddq)
+
+    @abstractmethod
+    def _check_single_state(self, task: Optional[Task.Task], robot: RobotBase, q: np.ndarray, dq: np.ndarray,
+                            ddq: Optional[np.ndarray] = None) -> bool:
+        """
+        This method needs to be implemented by all constraints that inherit from this mixin.
+        """
+
+
+class AlwaysTrueConstraint(ConstraintBase, RobotConstraint):
     """
     This constraint is always true and can be used for constraints ensured by the software library before evaluation.
     """
 
+    def _check_single_state(self, task: Optional[Task.Task], robot: RobotBase, q: np.ndarray, dq: np.ndarray,
+                            ddq: Optional[np.ndarray]) -> bool:
+        """As the name says, this constraint is always valid"""
+        return True
+
+    def is_valid_at(self, solution: Solution.SolutionBase, t: float) -> bool:
+        """As the name says, this constraint is always valid"""
+        return True
+
     def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
         """As the name says, this constraint is always valid"""
         return True
@@ -113,6 +180,11 @@ class AlwaysTrueConstraint(ConstraintBase):
         logging.info("Always true constraint is skipped during serialization")
         return {}
 
+    @property
+    def _equality_parameters(self) -> Tuple[any, ...]:
+        """Two always true constraints are always equal"""
+        return tuple()
+
 
 class GoalsBySameRobot(ConstraintBase):
     """Checks that all goals specified were achieved by the same robot."""
@@ -126,21 +198,17 @@ class GoalsBySameRobot(ConstraintBase):
         self.goals = set(goals)
         raise NotImplementedError("Goals by same robot not yet specified nor implemented")
 
+    def is_valid_at(self, solution: Solution.SolutionBase, t: float) -> bool:
+        """This constraint inherently needs access to more than one point in time"""
+        raise AttributeError("Goals by same robot constraint cannot be checked at one specific time")
+
     def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
         """Evaluates the solution up to time t"""
         # goal_ids, times = map(tuple, zip(*sorted(solution.t_goals.items(), key=lambda x: x[1])))
         # goals = [solution.task.goals_by_id[gid] for gid in goal_ids]
         # check_for = tuple(goal for i, goal in enumerate(goals) if goal.id in self.goals and times[t] <= t)
         # TODO: Check that all goals in check_for were done by the same robot
-        warn("Multi-Robot-Tasks not yet implemented", UserWarning)
-        return True
-
-    def fulfilled(self, solution: Solution.SolutionBase) -> bool:
-        """Checks the whole solution for all goals specified by this constraint"""
-        # goals = [goal for goal in solution.task.goals if goal.id in self.goals]
-        # TODO: Check that all goals with ID in self.goals were done by the same robot
-        warn("Multi-Robot-Tasks not yet implemented", UserWarning)
-        return True
+        raise NotImplementedError()
 
     def to_json_data(self) -> Dict[str, any]:
         """Dumps this constraint to a dictionary"""
@@ -149,6 +217,11 @@ class GoalsBySameRobot(ConstraintBase):
             'goals': list(self.goals)
         }
 
+    @property
+    def _equality_parameters(self) -> Tuple[str, ...]:
+        """Same goals, same constraint"""
+        return tuple(sorted(self.goals, key=lambda g: g.id))
+
 
 class GoalOrderConstraint(ConstraintBase):
     """Checks that the goals in the task are fulfilled in the right order."""
@@ -161,6 +234,10 @@ class GoalOrderConstraint(ConstraintBase):
         """
         self.order: Tuple[str, ...] = tuple(map(str, order))
 
+    def is_valid_at(self, solution: Solution.SolutionBase, t: float) -> bool:
+        """This constraint inherently needs access to more than one point in time"""
+        raise AttributeError("Goal order constraint cannot be checked at one specific time")
+
     def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
         """Evaluates the solution up to time t"""
         goal_ids, times = map(tuple, zip(*sorted(solution.t_goals.items(), key=lambda x: x[1])))
@@ -179,14 +256,27 @@ class GoalOrderConstraint(ConstraintBase):
             'order': self.order
         }
 
+    @property
+    def _equality_parameters(self) -> Tuple[str, ...]:
+        """Same order, same constraint"""
+        return tuple(self.order)
+
 
 class AllGoalsFulfilled(ConstraintBase):
     """Checks whether all goals are fulfilled"""
 
     global_only = True
 
+    def is_valid_at(self, solution: Solution.SolutionBase, t: float) -> bool:
+        """This constraint inherently needs access to more than one point in time"""
+        if t >= solution.time_steps[-1]:
+            return self.fulfilled(solution)
+        raise AttributeError("All goals fulfilled constraint cannot be checked at one specific time")
+
     def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
         """Test if all goals can still be fulfilled up to time t."""
+        if t >= solution.time_steps[-1]:
+            return self.fulfilled(solution)
         return True
 
     def fulfilled(self, solution: Solution.SolutionBase) -> bool:
@@ -197,8 +287,13 @@ class AllGoalsFulfilled(ConstraintBase):
         """Return json-able dict of this constraint."""
         return {'type': 'allGoalsFulfilled'}
 
+    @property
+    def _equality_parameters(self) -> Tuple[any, ...]:
+        """All goals, same constraint"""
+        return tuple()
 
-class JointLimits(ConstraintBase):
+
+class JointLimits(ConstraintBase, RobotConstraint):
     """Holds constraints on the robots (hardware) limits like position, velocity, torques, etc."""
 
     robot_limit_types = ('q', 'dq', 'ddq', 'tau')
@@ -220,26 +315,51 @@ class JointLimits(ConstraintBase):
         self.ddq = 'ddq' in parts
         self.tau = 'tau' in parts
 
-    def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
-        """Evaluates the solution up to time t"""
+    def _check_single_state(self, task: Task.Task, robot: RobotBase, q: np.ndarray, dq: np.ndarray,
+                            ddq: Optional[np.ndarray] = None) -> bool:
+        """
+        Checks whether the limits of the robot are held.
+
+        As a single state doesn't have a solution reference, this check assumes no external influences.
+        :param task: The task is not being used in this check.
+        :param robot: The robot for which the limits are checked.
+        :param q: The joint positions of the robot.
+        :param dq: The joint velocities of the robot.
+        :param ddq: The joint accelerations of the robot.
+        """
+        get_att = {'q': lambda: q, 'dq': lambda: dq, 'ddq': lambda: ddq if ddq is not None else np.zeros_like(q)}
+        get_att['tau'] = lambda: robot.id(q, dq, get_att['ddq']())
+        valid = True
+        for kind, robot_att in zip(self.robot_limit_types, self._robot_attributs):
+            if getattr(self, kind):  # Check what kinds of limits are tested
+                limits = getattr(robot, robot_att)  # Get the limits for the robot
+                state = get_att[kind]()
+                if len(limits.shape) == 2:  # Lower and upper limits
+                    valid = valid and np.all(limits[0, :] <= state) and np.all(limits[1, :] >= state)
+                else:
+                    valid = valid and np.all(np.abs(state) <= limits)
+        return valid
+
+    def is_valid_at(self, solution: Solution.SolutionBase, t: float) -> bool:
+        """Evaluates the solution at time t"""
         valid = True
         for kind, robot_att in zip(self.robot_limit_types, self._robot_attributs):
             if getattr(self, kind):  # Check what kinds of limits are tested
                 limits = getattr(solution.robot, robot_att)  # Get the limits for the robot
-                in_solution = getattr(solution, kind)[solution.get_time_id(t)]  # Get the real values up until t
+                in_solution = getattr(solution, kind)[solution.get_time_id(t)]  # Get the real values at time t
                 if len(limits.shape) == 2:  # Lower and upper limits
                     valid = valid and np.all(limits[0, :] <= in_solution) and np.all(limits[1, :] >= in_solution)
                 else:
                     valid = valid and np.all(np.abs(in_solution) <= limits)
         return valid
 
-    def fulfilled(self, solution: Solution.SolutionBase) -> bool:
-        """Checks the whole solution for all applied joint constraints"""
+    def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
+        """Evaluates the solution up to time t"""
         valid = True
         for kind, robot_att in zip(self.robot_limit_types, self._robot_attributs):
             if getattr(self, kind):  # Check what kinds of limits are tested
                 limits = getattr(solution.robot, robot_att)  # Get the limits for the robot
-                in_solution = getattr(solution, kind)  # Get the real values (in solution also called q, dq, ...)
+                in_solution = getattr(solution, kind)[:solution.get_time_id(t) + 1]  # Get the real values up until t
                 if len(limits.shape) == 2:  # Lower and upper limits
                     valid = valid and np.all(limits[0, :] <= in_solution) and np.all(limits[1, :] >= in_solution)
                 else:
@@ -253,6 +373,11 @@ class JointLimits(ConstraintBase):
             'parts': [kind for kind in self.robot_limit_types if getattr(self, kind)]
         }
 
+    @property
+    def _equality_parameters(self) -> Tuple[str, ...]:
+        """Same parts, same constraint"""
+        return tuple(kind for kind in self.robot_limit_types if getattr(self, kind))
+
 
 class BasePlacement(ConstraintBase):
     """Constraint on the placement of the base coordinate system."""
@@ -266,9 +391,12 @@ class BasePlacement(ConstraintBase):
         self.base_pose = base_pose
 
     @classmethod
-    def _from_json_data(cls, description: Dict[str, any]) -> ConstraintBase:
+    def _from_json_data(cls, description: Dict[str, any], *args, **kwargs) -> ConstraintBase:
         """Create Base placement constraint from json data."""
-        base_pose = ToleratedPose.from_json_data(description['pose'])
+        frame_name = 'constraint.base'
+        if 'frameName' in kwargs:
+            frame_name = f"{kwargs.pop('frameName', None)}.{frame_name}"
+        base_pose = ToleratedPose.from_json_data(description['pose'], **kwargs, frameName=frame_name)
         return cls(base_pose)
 
     @property
@@ -303,7 +431,7 @@ class BasePlacement(ConstraintBase):
 
         return sample
 
-    def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
+    def is_valid_at(self, solution: Solution.SolutionBase, t: float) -> bool:
         """We assume the base placement does not change, so this method makes no sense."""
         raise NotImplementedError("Assuming fixed base for now - use fulfilled method.")
 
@@ -319,12 +447,18 @@ class BasePlacement(ConstraintBase):
             "pose": self.base_pose.to_json_data()
         }
 
-    def visualize(self, viz: pinocchio.visualize.MeshcatVisualizer, scale: float = 1.) -> None:
+    def visualize(self, viz: pinocchio.visualize.MeshcatVisualizer, scale: float = 1., show_name: bool = False) -> None:
         """Visualize the constraint"""
-        self.base_pose.visualize(viz, scale=scale, name="base_constraint")
+        self.base_pose.visualize(viz, scale=scale, name="base_constraint", text="Base constraint" * show_name,
+                                 background_color="red", text_color="black")
+
+    @property
+    def _equality_parameters(self) -> Tuple[ToleratedPose]:
+        """Equal if the placements are equal"""
+        return (self.base_pose,)
 
 
-class CoterminalJointAngles(ConstraintBase):
+class CoterminalJointAngles(ConstraintBase, RobotConstraint):
     """Constrains a robot can reach a certain pose, being ignorant about full circle rotation deviations
 
     This constraint checks whether given joint position, given by joint angles, can be reached in general.
@@ -333,7 +467,7 @@ class CoterminalJointAngles(ConstraintBase):
 
     def __init__(self, limits: np.ndarray):
         """
-        :param limits: A 2xdof array of joint limits: [[upper_limits], [lower_limimts]]
+        :param limits: A 2xdof array of joint limits: [[lower_limits], [upper_limimts]]
         """
         if not limits.shape[0] == 2:
             raise ValueError("Joint limits must be given as array in (2, #dof) shape")
@@ -347,6 +481,11 @@ class CoterminalJointAngles(ConstraintBase):
         """Maps the input q to the range [-pi, pi)."""
         return np.mod(q + np.pi, 2 * np.pi) - np.pi
 
+    def _check_single_state(self, task: Optional[Task.Task], robot: RobotBase, q: np.ndarray,
+                            dq: np.ndarray, ddq: Optional[np.ndarray] = None) -> bool:
+        """Checks if q is within the joint limits, ignoring all other arguments."""
+        return self._validate(q)
+
     def _validate(self, q: np.ndarray) -> bool:
         """
         Returns whether q is within the joint limits
@@ -356,26 +495,27 @@ class CoterminalJointAngles(ConstraintBase):
         q = self.map(q)
         return all(self.limits[0, :] <= q) and all(self.limits[1, :] >= q)
 
-    def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
+    def is_valid_at(self, solution: Solution.SolutionBase, t: float) -> bool:
         """Checks whether the given joint angles are within the defined limits at time t."""
         q = solution.q[solution.get_time_id(t)]
         return self._validate(q)
 
-    def fulfilled(self, solution: Solution.SolutionBase) -> bool:
-        """Checks whether the given joint angles are within the defined limits in the solution."""
-        return all(self._validate(q) for q in solution.q)
-
     def to_json_data(self) -> Dict[str, any]:
         """Dumps this constraint to a dictionary"""
         raise NotImplementedError("Coterminal Joint Angles not yet available.")
 
+    @property
+    def _equality_parameters(self) -> Tuple[float, ...]:
+        """Flatten the limits to a 1d array"""
+        return tuple(self.limits.flatten())
+
 
 class JointAngles(CoterminalJointAngles):
     """This is a stronger constraint than coterminal joint angles, verifying exact angles against joint limits."""
 
     def __init__(self, limits: np.ndarray):
         """
-        limits: A 2xdof array of joint limits: [[upper_limits], [lower_limimts]]
+        limits: A 2xdof array of joint limits: [[lower_limits], [upper_limimts]]
         """
         super().__init__(limits)
 
@@ -389,38 +529,67 @@ class JointAngles(CoterminalJointAngles):
         raise NotImplementedError("Use JointAnglesRobot - independent joint angles constraint not yet defined")
 
 
-class CollisionFree(ConstraintBase):
+class CollisionFree(ConstraintBase, RobotConstraint):
     """A constraint that checks whether the robot is in collision with the environment or itself."""
 
-    def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
-        """Checks whether the robot is in collision with the environment or itself at time t."""
-        q = solution.q[solution.get_time_id(t)]
-        solution.robot.update_configuration(q)
-        return not solution.robot.has_collisions(solution.task)
+    def __init__(self, safety_margin: float = 0):
+        """
+        Initialize collision-free constraint
 
-    def fulfilled(self, solution: Solution.SolutionBase) -> bool:
-        """Evaluates the whole solution for collision-freeness."""
-        return all(self.is_valid_until(solution, t) for t in solution.time_steps)
+        :param safety_margin: A safety margin to use for collision checking. The default is 0, which means that the
+            collision checking is done with the exact geometries. For any value larger than 0, the collision checking
+            is done with the inflated geometries.
+        """
+        super().__init__()
+        if safety_margin < 0:
+            raise ValueError("Safety margin must be non-negative.")
+        self.safety_margin: float = safety_margin
+
+    def _check_single_state(self, task: Optional[Task.Task], robot: RobotBase, q: np.ndarray, dq: np.ndarray,
+                            ddq: Optional[np.ndarray] = None) -> bool:
+        """Checks the constraint for the given task in configuration q."""
+        robot.update_configuration(q)
+        return not robot.has_collisions(task, safety_margin=self.safety_margin)
+
+    def is_valid_at(self, solution: Solution.SolutionBase, t: float) -> bool:
+        """Checks whether the robot is in collision with the environment or itself at time t."""
+        return self._check_single_state(solution.task, solution.robot, solution.q[solution.get_time_id(t)],
+                                        solution.dq[solution.get_time_id(t)])
 
     def to_json_data(self) -> Dict:
         """Dumps this constraint to a dictionary"""
         return {'type': 'collisionFree'}
 
+    @property
+    def _equality_parameters(self) -> Tuple[any, ...]:
+        """Collision-free is equal to itself"""
+        return ()
 
-class SelfCollisionFree(CollisionFree):
+
+class SelfCollisionFree(CollisionFree, RobotConstraint):
     """A constraint that checks whether the robot is in collision with itself."""
 
-    def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
-        """Checks whether the robot is in collision with itself at time t."""
-        q = solution.q[solution.get_time_id(t)]
-        return not solution.robot.has_self_collision(q)
+    def __init__(self, safety_margin: float = 0):
+        """
+        Initialize self-collision-free constraint
+
+        :param safety_margin: For self collisions, this parameter is currently ignored.
+        """
+        super().__init__(safety_margin=safety_margin)
+        if safety_margin != 0:
+            logging.warning(f"The safety margin of {safety_margin} is currently ignored for self collision checking.")
+
+    def _check_single_state(self, task: Optional[Task.Task], robot: RobotBase, q: np.ndarray, dq: np.ndarray,
+                            ddq: Optional[np.ndarray] = None) -> bool:
+        """Checks the constraint for the given task in configuration q."""
+        return not robot.has_self_collision(q)
 
     def to_json_data(self) -> Dict[str, any]:
         """Dumps this constraint to a dictionary"""
         return {'type': 'selfCollisionFree'}
 
 
-class EndEffector(ConstraintBase):
+class EndEffector(ConstraintBase, RobotConstraint):
     """A constraint that checks whether the end-effector (eef) keeps a certain pose and/or velocity."""
 
     default_velocity_limits = np.asarray((-math.inf, math.inf))
@@ -438,9 +607,10 @@ class EndEffector(ConstraintBase):
         :param rotation_velocity_lim: Minimal / Maximal rotation velocity :math:`\omega` of the end-effector in
           :math:`\frac{rad}{s}`; defaults to any allowed
         """
-        self.pose = pose
-        self.velocity_lim = self.default_velocity_limits.copy() if velocity_lim is None else velocity_lim
-        self.rotation_velocity_lim = self.default_velocity_limits.copy() \
+        self.pose: Optional[ToleratedPose] = pose
+        self.velocity_lim: Optional[np.ndarray] = self.default_velocity_limits.copy() \
+            if velocity_lim is None else velocity_lim
+        self.rotation_velocity_lim: Optional[np.ndarray] = self.default_velocity_limits.copy() \
             if rotation_velocity_lim is None else rotation_velocity_lim
         if self.velocity_lim[0] > self.velocity_lim[1] or self.rotation_velocity_lim[0] > self.rotation_velocity_lim[1]:
             logging.warning("velocity_lim_min > velocity_lim_max or "
@@ -448,33 +618,35 @@ class EndEffector(ConstraintBase):
                             "eef_constraint will always be invalid.")
 
     @classmethod
-    def _from_json_data(cls, description: Dict[str, any]) -> ConstraintBase:
-        pose = ToleratedPose.from_json_data(description.pop('pose')) if "pose" in description else None
+    def _from_json_data(cls, description: Dict[str, any], *args, **kwargs) -> ConstraintBase:
+        frame_name = 'constraint.eef'
+        if 'frameName' in kwargs:
+            frame_name = f"{kwargs.pop('frameName', None)}.{frame_name}"
+        pose = ToleratedPose.from_json_data(description.pop('pose'), frameName=frame_name, **kwargs) \
+            if "pose" in description else None
         velocity_lim = np.asarray((description.pop('v_min', -math.inf), description.pop('v_max', math.inf)))
         rotation_velocity_lim = np.asarray((description.pop('o_min', -math.inf), description.pop('o_max', math.inf)))
         return cls(pose=pose, velocity_lim=velocity_lim, rotation_velocity_lim=rotation_velocity_lim)
 
-    def is_valid_until(self, solution: Solution.SolutionBase, t: float) -> bool:
-        """Checks whether the robot obeys eef velocity and pose constraints at time t."""
-        q = solution.q[solution.get_time_id(t)]
-        dq = solution.dq[solution.get_time_id(t)]
-
-        # Check eef velocities
-        solution.robot.update_configuration(q, dq)
-        v_robot = solution.robot.tcp_velocity
+    def _check_single_state(self, task: Task, robot: RobotBase, q: np.ndarray, dq: np.ndarray,
+                            ddq: Optional[np.ndarray] = None) -> bool:
+        """Can be used to check this constraint against a single, existing robot state. The task is ignored."""
+        robot.update_configuration(q, dq)
+        v_robot = robot.tcp_velocity
         v_robot, o_robot = np.linalg.norm(v_robot[:3]), np.linalg.norm(v_robot[3:])
         if v_robot < self.velocity_lim[0] or self.velocity_lim[1] < v_robot or \
                 o_robot < self.rotation_velocity_lim[0] or self.rotation_velocity_lim[1] < o_robot:
             return False
-
         # Check eef pose
         if self.pose is None:
             return True
-        return self.pose.valid(solution.robot.fk())
+        return self.pose.valid(robot.fk())
 
-    def fulfilled(self, solution: Solution.SolutionBase) -> bool:
-        """Evaluates the whole solution for obeying the eef constraint."""
-        return all(self.is_valid_until(solution, t) for t in solution.time_steps)
+    def is_valid_at(self, solution: Solution.SolutionBase, t: float) -> bool:
+        """Checks whether the robot obeys eef velocity and pose constraints at time t."""
+        q = solution.q[solution.get_time_id(t)]
+        dq = solution.dq[solution.get_time_id(t)]
+        return self.check_single_state(task=None, robot=solution.robot, q=q, dq=dq)
 
     def to_json_data(self) -> Dict[str, any]:
         """Dumps this constraint to a dictionary."""
@@ -486,3 +658,13 @@ class EndEffector(ConstraintBase):
         if self.pose is not None:
             data['pose'] = self.pose.to_json_data()
         return data
+
+    @property
+    def _equality_parameters(self) -> Tuple[any, ...]:
+        """End-effector constraints are equal if they have the same pose and limits"""
+        return (self.pose,) + tuple(self.velocity_lim.flat) + tuple(self.rotation_velocity_lim.flat)
+
+
+DEFAULT_CONSTRAINTS = (JointLimits(('q', 'dq', 'tau')),
+                       CollisionFree(),
+                       AllGoalsFulfilled())
