@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 # Author: Jonathan Külz
 # Date: 17.01.22
-from typing import Collection
+from typing import Collection, Optional
 
 import hppfcl
 import numpy as np
@@ -120,12 +120,14 @@ def axis_angle2rot_mat(axis_angle: np.ndarray) -> np.ndarray:
 
 def axis_angle_rotation(vec: np.ndarray, axis_angle: np.ndarray) -> np.ndarray:
     """
-    Rotates a vector by a given axis-angle representation.
+    Rotates a 3x1 vector by a given axis-angle representation.
 
     Equivalent to spatial.axis_angle2rot_mat(axis_angle) @ vec, but computationally more efficient.
+    :param vec: The vector to be rotated
+    :param axis_angle: The axis-angle representation of the rotation, either 4x1 or 3x1 array
     :source: https://en.wikipedia.org/wiki/Axis%E2%80%93angle_representation
     """
-    if vec.squeeze().shape == (3,):
+    if axis_angle.squeeze().shape == (3,):
         theta = np.linalg.norm(axis_angle)
         axis = axis_angle / theta
     else:
@@ -214,9 +216,11 @@ def mat2euler(R: np.ndarray, seq: str = 'xyz') -> np.ndarray:
     return Rotation.from_matrix(R.copy()).as_euler(seq)  # Calling copy to being able working with read-only arrays
 
 
-def random_rotation() -> np.ndarray:
+def random_rotation(rng: Optional[np.random.Generator] = None) -> np.ndarray:
     """Returns a random 3x3 rotation matrix."""
-    return Rotation.random().as_matrix()
+    if rng is not None:
+        rng = np.random.default_rng()
+    return Rotation.random(random_state=rng).as_matrix()
 
 
 def rotation_in_bounds(rot: np.ndarray, bounds: np.ndarray) -> np.ndarray:
