@@ -57,7 +57,7 @@ class TestModule(unittest.TestCase):
         # When adding the connectors, strict mode should work
         b2.connectors.add(c1)
         b3.connectors.add(c2)
-        joint_module._module_graph = joint_module._build_module_graph()  # Fix of hacking new connectors in module
+        joint_module._module_graph = joint_module.build_module_graph()  # Fix of hacking new connectors in module
         self.assertIs(get_module_type(joint_module), ModuleType.JOINT)
 
         base_connector = Bodies.Connector('BC', parent=b4, connector_type='base')
@@ -183,6 +183,9 @@ class TestModule(unittest.TestCase):
         same_module = AtomicModule(header_one, [generic_body, box_body, another_body], [jnt, jnt_works])
 
         self.assertEqual(module, same_module)
+        self.assertEqual(hash(module), hash(same_module))
+        self.assertEqual(module.num_connectors,
+                         len(generic_body.connectors.union(box_body.connectors).union(another_body.connectors)))
         self.assertEqual(module.mass, generic_body.mass + box_body.mass + another_body.mass)
 
         same_module.bodies.pop()  # Don't do this at home
@@ -297,13 +300,21 @@ class TestModule(unittest.TestCase):
         """Check parameterized modules that contain at least one joint"""
         straight_header = ModuleHeader('J1', 'Straight Prismatic Joint')
         straight = ParameterizedStraightJoint(straight_header, radius=.1, joint_type=TimorJointType.prismatic,
+                                              limits=((0, 1), (0, 20)),
                                               joint_parameters={'q_limits': (-.3, .5)})
         orthogonal_header = ModuleHeader('J2', 'Orthogonal Revolute Joint')
-        orthogonal = ParameterizedOrthogonalJoint(orthogonal_header, radius=.1, joint_type=TimorJointType.revolute)
+        orthogonal = ParameterizedOrthogonalJoint(orthogonal_header, radius=.1, joint_type=TimorJointType.revolute,
+                                                  limits=((0, 1), (0, 10), (0, 9.5)))
 
         self.assertEqual(straight.joint.type, TimorJointType.prismatic)
         self.assertEqual(orthogonal.joint.type, TimorJointType.revolute)
 
+        for module in (straight, orthogonal):
+            cpy = module.copy('cpy')
+            self.assertEqual(module.parameters, cpy.parameters)
+            for b1, b2 in zip(sorted(module.bodies, key=lambda x: x._id), sorted(cpy.bodies, key=lambda x: x._id)):
+                self.assertTrue(np.all(b1.parameter_limits == b2.parameter_limits))
+
         Gs = straight.module_graph
         self.assertTrue(nx.is_weakly_connected(Gs))
         nodes = {n._id: n for n in Gs.nodes}
