@@ -7,21 +7,29 @@ import pinocchio as pin
 
 from timor.Module import ModuleAssembly, ModulesDB
 from timor.Robot import PinRobot, RobotBase
-from timor.task import Tolerance
+from timor.task import Constraints, Tolerance
 from timor.utilities import logging, prebuilt_robots, spatial
 from timor.utilities.file_locations import robots
+from timor.utilities.frames import Frame, NOMINAL_FRAME, WORLD_FRAME
 from timor.utilities.tolerated_pose import ToleratedPose
 from timor.utilities.transformation import Transformation
 
 
 class PinocchioRobotSetup(unittest.TestCase):
-    """Test if robots can instantiated, and works as expected."""
+    """Test if robots can be instantiated, and works as expected."""
+
+    @staticmethod
+    def get_random_config(rng: np.random.Generator, robot: RobotBase):
+        """Random configuration generator depending on numpy only (no pinocchio, so randomness is controlled)"""
+        lim_lower = np.maximum(robot.joint_limits[0, :], -2 * np.pi)
+        lim_upper = np.minimum(robot.joint_limits[1, :], 2 * np.pi)
+        return (lim_upper - lim_lower) * rng.random(robot.dof) + lim_lower
 
     def setUp(self) -> None:
         self.package_dir = robots['panda'].parent
         self.urdf = robots['panda'].joinpath('urdf').joinpath('panda.urdf')
         random.seed(1234)
-        np.random.seed(1234)
+        self.rng = np.random.default_rng(1234)
 
     def test__first_load_robot(self):
         """Double underscore because tests are run alphabetically and this one should run first"""
@@ -30,14 +38,13 @@ class PinocchioRobotSetup(unittest.TestCase):
         self.assertIsInstance(robot, RobotBase)
 
     def test_base_placement(self):
-        np.random.seed(99)
         new_placement = Transformation.random()
 
         # Check that moving the robot to a new placement works
         robot_one = PinRobot.from_urdf(self.urdf, self.package_dir)
         robot_two = PinRobot.from_urdf(self.urdf, self.package_dir, base_placement=new_placement)
         robot_one.move(new_placement)
-        np_test.assert_array_equal(robot_one.placement, robot_two.placement)
+        np_test.assert_array_equal(robot_one.placement.homogeneous, robot_two.placement.homogeneous)
 
         for _ in range(5):
             robot_one = PinRobot.from_urdf(self.urdf, self.package_dir)
@@ -91,7 +98,7 @@ class PinocchioRobotSetup(unittest.TestCase):
 
         # Velocity should be close to omega x r if single joint moves
         for _ in range(100):
-            q = robot.random_configuration()
+            q = self.get_random_config(self.rng, robot)
             # First joint movement. Needs relative movement of 5th (= 1st joint) and EEF frame; specific to Panda robot
             robot.update_configuration(q, dq0)
             fks = robot.fk(kind="full")
@@ -112,7 +119,7 @@ class PinocchioRobotSetup(unittest.TestCase):
             np_test.assert_array_almost_equal(v_rob[3:], fks[7].rotation @ np.asarray((0., 0., 1.)))
 
             # Velocity should not change with translation of base
-            robot.set_base_placement(Transformation.from_translation(np.random.random((3,))))
+            robot.set_base_placement(Transformation.from_translation(self.rng.random((3,))))
             robot.update_configuration(q, dq1)
             v_rob_moved = robot.tcp_velocity
             np_test.assert_array_almost_equal(v_rob, v_rob_moved)
@@ -128,11 +135,68 @@ class PinocchioRobotSetup(unittest.TestCase):
             # reset
             robot.set_base_placement(Transformation.neutral())
 
+    def test_jacobian_ik_masked(self):
+        robot = PinRobot.from_urdf(self.urdf, self.package_dir, rng=self.rng)
+        # Create some goals that are not reachable and then try to optimize for certain error terms only:
+        N = 50
+        fails = 0
+
+        def cost_rot(_r: RobotBase, _q: np.ndarray, _g: Transformation):
+            """A custom cost function to optimize for rotation only"""
+            return _r.fk(_q, 'tcp').distance(_g).rotation_angle
+
+        def cost_trans(_r: RobotBase, _q: np.ndarray, _g: Transformation):
+            """A custom cost function to optimize for translation only"""
+            return _r.fk(_q, 'tcp').distance(_g).translation_euclidean
+
+        def get_trans_z_cost(f: str):
+            def cost_trans_z(_r: RobotBase, _q: np.ndarray, _g: Transformation):
+                """A custom cost function to optimize for translation in z-direction only"""
+                if f == 'tcp':
+                    ref = NOMINAL_FRAME
+                else:
+                    ref = WORLD_FRAME
+                error = _g.inv @ _r.fk(_q)
+                return np.abs(ref.rotate_to_this_frame(error, _g).translation[2])
+            return cost_trans_z
+
+        trans_mask = np.array([True, True, True, False, False, False])
+        rot_mask = np.array([False, False, False, True, True, True])
+        z_mask = np.array([False, False, True, False, False, False])
+
+        def make_distance_matrix(q_s, costs, _g):
+            """Compute distance matrix for a set of joint configurations and a cost function"""
+            return np.asarray([[c(robot, q, _g.nominal.in_world_coordinates()) for q in q_s] for c in costs])
+
+        for _ in range(N):
+            for frame in ('tcp', 'world'):
+                goal = ToleratedPose(Transformation.random(self.rng) @
+                                     Transformation.from_translation(self.rng.random((3,)) * 100))
+                q_init = robot.random_configuration(self.rng)
+                kwargs = {'eef_pose': goal, 'ik_method': 'jacobian', 'max_iter': 50, 'allow_random_restart': True,
+                          'tcp_or_world_frame': frame, 'q_init': q_init}
+                q_rot, s1 = robot.ik(ik_cost_function=cost_rot, error_term_mask=rot_mask, **kwargs)
+                q_trans, s2 = robot.ik(ik_cost_function=cost_trans, error_term_mask=trans_mask, **kwargs)
+                q_trans_z, s3 = robot.ik(ik_cost_function=get_trans_z_cost(frame), error_term_mask=z_mask, **kwargs)
+
+                # If the goal was reachable, the error masking could be without effect
+                self.assertFalse(s1 or s2 or s3)
+
+                mat = make_distance_matrix([q_rot, q_trans, q_trans_z, q_init],
+                                           [cost_rot, cost_trans, get_trans_z_cost(frame)], goal)
+
+                self.assertEqual(np.argmin(mat[0, :]), 0)
+                fails += np.min(mat[1, :]) != mat[1, 1]  # numeric IK is not super precise, fails _can_ happen
+                # Minimizing the z-error in tcp frame is actually not trivial, but we can test for the world frame
+                if frame == 'world':
+                    self.assertLess(mat[2, 2], mat[2, 3])
+            self.assertLess(fails, int(0.1 * (N * 2)), msg=f"Failed to optimize for position only {fails} times")
+
     def test_robot_fk(self):
         robot = PinRobot.from_urdf(self.urdf, self.package_dir)
         self.assertIsInstance(robot.fk(robot.configuration, 'tcp'), Transformation)
 
-        conf = pin.randomConfiguration(robot.model)
+        conf = self.get_random_config(self.rng, robot)
         for kind in ('tcp', 'joints', 'full'):
             # Updating geometry placements should not change kinematics
             self.assertEqual(
@@ -154,18 +218,15 @@ class PinocchioRobotSetup(unittest.TestCase):
 
         def random_values(limits: np.ndarray, N: int = 100):
             """Expects positive limits end returns random values between -limits and +limits"""
-            sign_mask = (np.random.random([N, len(limits)]) > .5)
-            rand_val = np.random.random([N, len(limits)]) * limits
+            sign_mask = (self.rng.random([N, len(limits)]) > .5)
+            rand_val = self.rng.random([N, len(limits)]) * limits
             rand_val[sign_mask] = rand_val[sign_mask] * -1
             return rand_val
 
-        q = np.vstack([pin.randomConfiguration(robot.model) for _ in range(100)])
+        q = np.vstack([self.get_random_config(self.rng, robot) for _ in range(100)])
         dq = random_values(robot.joint_velocity_limits, 100)
         tau = random_values(robot.model.effortLimit, 100)
 
-        random.seed(1001)
-        np.random.seed(1001)
-
         for i in range(q.shape[0]):
             # Only considers robot bodies; ignores motor dynamics and friction
             ddq = pin.aba(robot.model, robot.data, q[i], dq[i], tau[i])
@@ -193,7 +254,7 @@ class PinocchioRobotSetup(unittest.TestCase):
             return _r.fk(_q, 'tcp').distance(_g).translation_euclidean
 
         error_count = {"scipy": 0, "jacobian": 0}
-        N = 100
+        N = 50
         n = 0
         scipy_tolerance = Tolerance.CartesianXYZ.default()
         jacobian_tolerance = Tolerance.DEFAULT_SPATIAL
@@ -202,28 +263,42 @@ class PinocchioRobotSetup(unittest.TestCase):
         zero_limits = np.zeros_like(robot.joint_torque_limits)
         for _ in range(N):
             # Using a random example for which we know, there's at least one goal configuration
-            conf = pin.randomConfiguration(robot.model)
-            goal = robot.fk(conf)
+            conf = self.get_random_config(self.rng, robot)
+            q_init = self.get_random_config(self.rng, robot)
+            goal = Frame("", robot.fk(conf), WORLD_FRAME)
             if robot.has_self_collision():
                 # Don't test ik in these cases - it might fail due to avoiding self collisions
                 continue
             n += 1  # Keep track of how many examples were really worked with
             robot.model.effortLimit = inf_limits  # Make it impossible to fail due to the torque limits
-            robot.update_configuration(pin.neutral(robot.model))
-            conf, success = robot.ik_scipy(ToleratedPose(goal, scipy_tolerance), ik_cost_function=translation_cost)
+            robot.update_configuration(q_init)
+            conf, success = robot.ik(ToleratedPose(goal, scipy_tolerance), q_init=q_init, ik_method='scipy',
+                                     ik_cost_function=translation_cost)
+            # Check that multiple runs produce different ik solutions
+            q_s = [robot.ik(ToleratedPose(goal, scipy_tolerance), ik_cost_function=translation_cost, ik_method='scipy')
+                   for _ in range(10)]
+            q_s = np.asarray([q for q, s in q_s if s])
+            dist = np.linalg.norm(q_s - q_s[0, :], axis=1)
+            self.assertTrue(np.any(dist > 1e-3))
             if success:
-                # ik_scipy optimizes only for position, therefore the Cartesian XYZ tolerance
-                self.assertTrue(scipy_tolerance.valid(goal, robot.fk(conf)))
+                # Here, we only optimize for position, therefore the Cartesian XYZ tolerance
+                self.assertTrue(scipy_tolerance.valid(goal.in_world_coordinates(), robot.fk(conf)))
                 self.assertTrue(max(abs(conf)) < 2 * np.pi)  # IK result should be mapped to (-2pi, 2pi)
                 self.assertFalse(robot.has_self_collision(conf))
             else:
                 error_count['scipy'] += 1
 
-            robot.update_configuration(pin.neutral(robot.model))  # Prevent using the scipy result as initial guess
-            conf, success = robot.ik_jacobian(ToleratedPose(goal, jacobian_tolerance))
+            robot.update_configuration(q_init)  # Prevent using the scipy result as initial guess
+            conf, success = robot.ik(ToleratedPose(goal, jacobian_tolerance), ik_method='jacobian')
+            # Check that multiple runs produce different ik solutions
+            q_s = [robot.ik(ToleratedPose(goal, jacobian_tolerance), ik_method='jacobian') for _ in range(10)]
+            q_s = np.asarray([q for q, s in q_s if s])
+            dist = np.linalg.norm(q_s - q_s[0, :], axis=1)
+            self.assertTrue(np.any(dist > 1e-3))
             if success:
-                np_test.assert_array_almost_equal(goal.homogeneous, robot.fk(conf).homogeneous, decimal=2)
-                self.assertTrue(jacobian_tolerance.valid(goal, robot.fk(conf)))
+                np_test.assert_array_almost_equal(goal.in_world_coordinates().homogeneous, robot.fk(conf).homogeneous,
+                                                  decimal=2)
+                self.assertTrue(jacobian_tolerance.valid(goal.in_world_coordinates(), robot.fk(conf)))
                 self.assertTrue(max(abs(conf)) < 2 * np.pi)
                 self.assertFalse(robot.has_self_collision(conf))
             else:
@@ -231,24 +306,27 @@ class PinocchioRobotSetup(unittest.TestCase):
                 fails.append((goal, conf))
 
             # Test that masking joints in the ik works
-            q_start = robot.random_configuration()
+            q_start = self.get_random_config(self.rng, robot)
             mask = np.array([random.random() > 0.5 for _ in range(len(conf))])
-            conf, success = robot.ik_jacobian(ToleratedPose(goal, jacobian_tolerance), q_init=q_start, joint_mask=mask)
+            conf, success = robot.ik(ToleratedPose(goal, jacobian_tolerance), q_init=q_start, joint_mask=mask,
+                                     ik_method='jacobian')
             np_test.assert_array_almost_equal(q_start[~mask], conf[~mask])
-            conf, success = robot.ik_scipy(ToleratedPose(goal, jacobian_tolerance), q_init=q_start, joint_mask=mask)
+            conf, success = robot.ik(ToleratedPose(goal, jacobian_tolerance), q_init=q_start,
+                                     ik_method='scipy', joint_mask=mask)
             np_test.assert_array_almost_equal(q_start[~mask], conf[~mask])
 
             # Test joint torque limits
             robot.model.effortLimit = zero_limits
-            conf, success = robot.ik_jacobian(ToleratedPose(goal, jacobian_tolerance), max_iter=50,
-                                              check_static_torques=True)
+            torque_constraint = Constraints.JointLimits(parts=('tau',))
+            conf, success = robot.ik(ToleratedPose(goal, jacobian_tolerance), max_iter=50, ik_method='jacobian',
+                                     outer_constraints=(torque_constraint,))
             self.assertFalse(success)
-            conf, success = robot.ik_scipy(ToleratedPose(goal, jacobian_tolerance), max_iter=50,
-                                           check_static_torques=True)
+            conf, success = robot.ik(ToleratedPose(goal, jacobian_tolerance), max_iter=50, ik_method='scipy',
+                                     outer_constraints=(torque_constraint,))
             self.assertFalse(success)
 
         for k, v in error_count.items():
-            self.assertLess(v, 0.05 * n, f"ik should be able to resolve almost any fk; "
+            self.assertLess(v, 0.03 * n, f"ik should be able to resolve almost any fk; "
                                          f"not for {k} with {v / n * 100:.2f}% errors.")
 
         if error_count['scipy'] + error_count['jacobian'] > 0:
@@ -264,7 +342,7 @@ class PinocchioRobotSetup(unittest.TestCase):
         wrapper_two = pin.RobotWrapper.BuildFromURDF(str(self.urdf), str(self.package_dir))
         robot_default = PinRobot(wrapper=wrapper_one)
         robot_moved = PinRobot(wrapper=wrapper_two, base_placement=displacement.copy())
-        conf = pin.randomConfiguration(robot_default.model)
+        conf = self.get_random_config(self.rng, robot_default)
 
         self.assertEqual(
             tuple(p.multiply_from_left(displacement) for p in robot_default.fk(conf, 'full')[1:]),
