@@ -1,8 +1,6 @@
 #!/usr/bin/env python3
 # Author: Jonathan Külz
 # Date: 17.02.22
-import math
-import random
 import unittest
 
 import numpy as np
@@ -14,8 +12,9 @@ from timor.task import Constraints, CostFunctions, Goals, Solution, Task, Tolera
 from timor.task.CostFunctions import QDist
 from timor.task.Solution import SolutionHeader, SolutionTrajectory
 from timor.task.Task import TaskHeader
-from timor.utilities import prebuilt_robots, spatial, visualization
+from timor.utilities import prebuilt_robots, visualization
 from timor.utilities.file_locations import get_test_tasks, robots
+from timor.utilities.frames import Frame, WORLD_FRAME
 from timor.utilities.prebuilt_robots import get_six_axis_assembly
 from timor.utilities.tolerated_pose import ToleratedPose
 from timor.utilities.trajectory import Trajectory
@@ -39,72 +38,8 @@ class TestSolution(unittest.TestCase):
         self.task_dir = file_locations["task_dir"]
         self.asset_dir = file_locations["asset_dir"]
 
-        random.seed(123)
-        np.random.seed(123)
-        self.random_configurations = np.asarray([pin.randomConfiguration(self.robot.model) for _ in range(200)])
-
-    def test_constraint_abstraction_level(self):
-        panda = Robot.PinRobot.from_urdf(self.panda_urdf, robots['panda'].parent)
-
-        limits = panda.joint_limits
-        base_constraint = Constraints.BasePlacement(ToleratedPose(Transformation.neutral(),
-                                                                  tolerance=Tolerance.DEFAULT_SPATIAL))
-        basic_joint_constraint = Constraints.JointLimits(parts=('q',))
-        additional_joint_constraint = Constraints.CoterminalJointAngles(limits * 0.95)
-
-        start_conf = np.mean(limits * np.random.random(limits.shape), axis=0)  # Around middle of limits
-        end_conf = np.mean(limits * np.random.random(limits.shape), axis=0)  # Around middle of limits
-
-        goal_pose = ToleratedPose(panda.fk(end_conf))
-        self.assertFalse(panda.has_self_collision(end_conf))
-        at_goal = Goals.At('At_Goal', goal_pose, constraints=[additional_joint_constraint])
-
-        task = Task.Task(self.task_header, goals=[at_goal],
-                         constraints=[base_constraint, basic_joint_constraint, Constraints.AllGoalsFulfilled()])
-        task_without_all_goals_fulfilled = Task.Task(self.task_header, goals=[at_goal],
-                                                     constraints=[base_constraint, basic_joint_constraint])
-
-        steps = 100
-        dt = .1
-        easy_trajectory = Trajectory(t=dt, q=np.linspace(start_conf, end_conf, steps + 1),
-                                     goal2time={at_goal.id: steps * dt})
-        solution_header = Solution.SolutionHeader(task.id)
-        cost_function = CostFunctions.CycleTime()
-        solution = Solution.SolutionTrajectory(easy_trajectory, solution_header, task=task,
-                                               assembly=ModuleAssembly.from_monolithic_robot(panda),
-                                               cost_function=cost_function)
-        solution_without_all_goals_fulfilled = Solution.SolutionTrajectory(
-            easy_trajectory, solution_header, task=task_without_all_goals_fulfilled,
-            assembly=ModuleAssembly.from_monolithic_robot(panda), cost_function=cost_function)
-
-        self.assertTrue(solution.valid)
-        solution._valid.value = None  # Dirty but quick reset of solution
-
-        # Test that local constraints can fail a solution
-        previous_limits = additional_joint_constraint.limits.copy()
-        additional_joint_constraint.limits = np.zeros_like(additional_joint_constraint.limits)
-        self.assertFalse(solution.valid)
-        # Should be valid as not every goal needs to be valid
-        self.assertTrue(solution_without_all_goals_fulfilled.valid)
-        solution._valid.value = None  # Dirty but quick reset of solution
-
-        # Reset to valid
-        additional_joint_constraint.limits = previous_limits
-
-        # Now hurt these local constraints, but outside the goal
-        addon_trajectory = Trajectory(t=dt, q=np.linspace(end_conf, panda.joint_limits[1, :], steps + 1))
-        solution.trajectory = solution.trajectory + addon_trajectory
-        self.assertTrue(solution.valid)
-        solution._valid.value = None  # Dirty but quick reset of solution
-
-        # But it fails if the same constraint is a global one
-        task.constraints = tuple(list(task.constraints) + [additional_joint_constraint])
-        self.assertFalse(solution.valid)
-
-        for global_constraint in (Constraints.GoalOrderConstraint(['At_Goal']),
-                                  base_constraint):
-            with self.assertRaises(ValueError):
-                Goals.At('At_Goal', goal_pose, constraints=[global_constraint])
+        self.rng = np.random.default_rng(123)
+        self.random_configurations = np.asarray([self.robot.random_configuration(self.rng) for _ in range(200)])
 
     def test_cost_parsing(self):
         """
@@ -114,15 +49,15 @@ class TestSolution(unittest.TestCase):
 
         for _ in range(100):
             # Generate random descriptor string
-            num_costs = random.randint(1, len(abbreviations))
-            num_weights = random.randint(1, len(abbreviations))
+            num_costs = self.rng.integers(1, len(abbreviations))
+            num_weights = self.rng.integers(1, len(abbreviations))
             cost_string = ''
             for i in range(num_costs):
                 if i >= num_weights:
-                    weight = str(float(random.random()))
-                    cost_string += random.choice(abbreviations) + '_' + weight
+                    weight = str(float(self.rng.random()))
+                    cost_string += self.rng.choice(abbreviations) + '_' + weight
                 else:
-                    cost_string += random.choice(abbreviations)
+                    cost_string += self.rng.choice(abbreviations)
                 if i < num_costs - 1:
                     cost_string += ','
 
@@ -156,91 +91,6 @@ class TestSolution(unittest.TestCase):
         with self.assertRaises(KeyError):
             CostFunctions.CostFunctionBase.from_descriptor('not_a_cost')
 
-    def test_eef_constraint(self):
-        eef_constraint_allow_any_pose = Constraints.EndEffector(
-            pose=ToleratedPose(nominal=Transformation.neutral(), tolerance=Tolerance.AlwaysValidTolerance()))
-        qs = np.asarray([self.robot.random_configuration() for _ in range(10)])
-        eef_constraint_allow_first_pose = Constraints.EndEffector(
-            pose=ToleratedPose(self.robot.fk(qs[0, :]), tolerance=Tolerance.DEFAULT_SPATIAL)
-        )
-
-        # Check pose
-        sol = Solution.SolutionTrajectory(Trajectory(t=1., q=qs, goal2time={}), Solution.SolutionHeader("tmp"),
-                                          Task.Task(Task.TaskHeader('tmp'), ),
-                                          ModuleAssembly.from_monolithic_robot(self.robot), CostFunctions.RobotMass(),
-                                          Transformation.neutral())
-        self.assertTrue(eef_constraint_allow_first_pose.is_valid_until(sol, 0.))
-        self.assertFalse(eef_constraint_allow_first_pose.is_valid_until(sol, 1.))
-        self.assertFalse(eef_constraint_allow_first_pose.fulfilled(sol))
-        self.assertTrue(eef_constraint_allow_any_pose.is_valid_until(sol, 0.))
-        self.assertTrue(eef_constraint_allow_any_pose.is_valid_until(sol, 1.))
-        self.assertTrue(eef_constraint_allow_any_pose.fulfilled(sol))
-
-        # Constructed case - movement along line with invariant rotation about line, e.g. drilling
-        orientation_tolerance = Tolerance.RotationAxisAngle(n_x=(-math.pi / 100, math.pi / 100),
-                                                            n_y=(-math.pi / 100, math.pi / 100),
-                                                            n_z=(-1, 1),  # Any rotation about local z-axis is ok
-                                                            theta_r=(0, np.pi))
-        ik_tolerance = Tolerance.Composed((orientation_tolerance, Tolerance.CartesianXYZ.default()))
-        start_pose = Transformation.from_translation((0.5, 0., 0.5)) @ spatial.rotY(math.pi / 2)
-        # Construct movement steps - q_start, q_end, q_mid_ok, q_mid_nok (small rot), q_end_to_far
-        qs = []
-        N_middle = 10  # How many random valid position inbetween start and end pose
-        t_good = 2 + N_middle  # start, end + valid in the middle
-        for T in (start_pose,  # q_start
-                  Transformation.from_translation((.2, 0., 0.)) @ start_pose,  # q_end
-                  # q_mid - all with random translation in 0, 0.2 and rotation about drill axis
-                  *(Transformation.from_translation((random.random() * 0.2, 0., 0.)) @
-                    start_pose @ spatial.rotZ(random.random()) for _ in range(N_middle)),
-                  Transformation.from_translation((.11, 0., 0.)) @ start_pose @ spatial.rotX(0.02),  # q_mid_not_ok
-                  Transformation.from_translation((.11, 0., 0.)) @ start_pose @ spatial.rotY(-0.02),  # q_mid_not_ok
-                  Transformation.from_translation((.22, 0., 0.)) @ start_pose  # q_end_to_far
-                  ):
-            q, success = self.robot.ik(ToleratedPose(T, ik_tolerance))
-            self.assertTrue(success, f"ik for {str(T)} failed")
-            qs.append(q)
-        drill_eef_constraint = Constraints.EndEffector(pose=ToleratedPose(
-            nominal=start_pose,
-            tolerance=Tolerance.Composed((
-                Tolerance.CartesianXYZ((-0.01, 0.01), (-0.01, 0.01), (-0.01, 0.2)),  # Drill along z
-                orientation_tolerance
-            ))
-        ))
-        drill_test_traj = Trajectory(1., q=np.asarray(qs), goal2time={})
-        sol = Solution.SolutionTrajectory(drill_test_traj, Solution.SolutionHeader("tmp"),
-                                          Task.Task(Task.TaskHeader("tmp")),
-                                          ModuleAssembly.from_monolithic_robot(self.robot), CostFunctions.RobotMass(),
-                                          Transformation.neutral())
-        for t in np.arange(0, t_good, 1):
-            self.assertTrue(drill_eef_constraint.is_valid_until(sol, t), f"Error at step {t}")
-        for t in np.arange(t_good, t_good + 3, 1):
-            self.assertFalse(drill_eef_constraint.is_valid_until(sol, t))
-
-        # Check velocities
-        self.robot.update_configuration(np.zeros((7,)))  # ~0.6 m from first joint to eef
-        eef_constraint_v = Constraints.EndEffector(velocity_lim=np.asarray((.55, 0.65)))
-        eef_constraint_o = Constraints.EndEffector(rotation_velocity_lim=np.asarray((.95, 1.05)))
-        vel_test_traj = Trajectory(t=np.asarray([0., 1., 2., 3., 4.]), q=np.zeros((5, 7)), goal2time={},
-                                   dq=np.asarray(((0., 1., 0., 0., 0., 0., 0.),  # eef ~ 0.6 m/s
-                                                  (0., .9, 0., 0., 0., 0., 0.),  # eef ~ 0.54 m/s - invalid
-                                                  (0., 1.1, 0., 0., 0., 0., 0.),  # eef ~ 0.66 m/s - invalid
-                                                  (0., -1., 0., 0., 0., 0., 0.),  # eef ~ 0.6 m/s
-                                                  (0., -.85, 0., 0., 0., 0., 0.),  # eef ~ 0.5 m/s
-                                                  )))
-        sol = Solution.SolutionTrajectory(vel_test_traj, Solution.SolutionHeader("tmp"),
-                                          Task.Task(Task.TaskHeader("tmp"), constraints=(eef_constraint_v,)),
-                                          ModuleAssembly.from_monolithic_robot(self.robot), CostFunctions.RobotMass(),
-                                          Transformation.neutral())
-        for c in (eef_constraint_v, eef_constraint_o):
-            self.assertTrue(c.is_valid_until(sol, 0.))
-            self.assertTrue(c.is_valid_until(sol, 3.))
-            self.assertFalse(c.is_valid_until(sol, 1.))
-            self.assertFalse(c.is_valid_until(sol, 2.))
-            self.assertFalse(c.is_valid_until(sol, 4.))
-        for t in np.arange(0., 3., 1.):
-            self.assertTrue(eef_constraint_allow_any_pose.is_valid_until(sol, t))
-            self.assertFalse(eef_constraint_allow_first_pose.is_valid_until(sol, t))
-
     def test_goals(self):
         task = Task.Task({'ID': 'dummy'})
         cost_function = CostFunctions.GoalsFulfilled()
@@ -249,20 +99,20 @@ class TestSolution(unittest.TestCase):
         pause_duration = .4
         dt = 0.01
         time_steps = 100
-        q_array_shape = (time_steps, robot.njoints)
+        q_array_shape = (time_steps, robot.dof)
         # Add constraint such that _valid can be tested as well
         pause_goal = Goals.Pause(ID='pause', duration=pause_duration,
                                  constraints=[Constraints.JointAngles(robot.joint_limits)])
 
         for _ in range(10):
-            bad_trajectory = Trajectory(t=dt, q=np.random.random(q_array_shape))
-            t_start = random.choice(tuple(bad_trajectory.t[bad_trajectory.t < max(bad_trajectory.t) - pause_duration]))
+            bad_trajectory = Trajectory(t=dt, q=self.rng.random(q_array_shape))
+            t_start = self.rng.choice(list(bad_trajectory.t[bad_trajectory.t < max(bad_trajectory.t) - pause_duration]))
             t_end = t_start + pause_duration
             q_good = bad_trajectory.q.copy()
 
             # Manually ensure there is a pause
             pause_mask = np.logical_and(t_start <= bad_trajectory.t, bad_trajectory.t <= t_end)
-            q_good[pause_mask] = random.random()
+            q_good[pause_mask] = self.rng.random()
             good_trajectory = Trajectory(t=dt, q=q_good)
             good_trajectory.dq[pause_mask] = 0
             good_trajectory.ddq[pause_mask] = 0
@@ -279,7 +129,11 @@ class TestSolution(unittest.TestCase):
 
             self.assertTrue(pause_goal.achieved(valid))
             self.assertFalse(pause_goal.achieved(invalid))
-            good_trajectory.dq = np.random.random(q_array_shape)
+            pause_goal.constraints.append(Constraints.AlwaysTrueConstraint())  # should not change anything
+            self.assertTrue(pause_goal.achieved(valid))
+            self.assertFalse(pause_goal.achieved(invalid))
+
+            good_trajectory.dq = self.rng.random(q_array_shape)
             self.assertFalse(pause_goal.achieved(valid))
 
             bad_trajectory_to_short = Trajectory.stationary(q=np.zeros(q_array_shape[1]),
@@ -291,15 +145,15 @@ class TestSolution(unittest.TestCase):
             self.assertFalse(pause_goal.achieved(invalid_too_short))
 
             # Test that times in goal duration rightly calculated
-            sol_times, sol_times_indices = pause_goal._get_time_range_goal(valid)
+            sol_times, sol_times_indices = pause_goal.get_time_range_goal(valid)
             self.assertListEqual(np.where(pause_mask)[0].tolist(), list(sol_times_indices))
             np_test.assert_array_less(t_start - dt, np.asarray(sol_times))
             np_test.assert_array_less(np.asarray(sol_times), t_end + dt)
 
             # Test constraint in duration - replace a single time-step inside pause with joint limit + a_bit_over_0
             q_bad_outside_joint_limits = good_trajectory.q.copy()
-            q_bad_outside_joint_limits[np.random.choice(np.argwhere(pause_mask).squeeze())] = \
-                robot.joint_limits[1, :] + np.random.random(robot.njoints)
+            q_bad_outside_joint_limits[self.rng.choice(np.argwhere(pause_mask).squeeze())] = \
+                robot.joint_limits[1, :] + self.rng.random(robot.dof)
             bad_trajectory_outside_joint_limits = Trajectory(t=dt, q=q_bad_outside_joint_limits,
                                                              goal2time=good_trajectory.goal2time)
             bad_sol_outside_joint_limits = \
@@ -313,15 +167,17 @@ class TestSolution(unittest.TestCase):
         # Set up a couple of realistic constraints
         constraint_joints = Constraints.JointAngles(self.robot.joint_limits)
         pos_tolerance = Tolerance.CartesianXYZ(*[[-.1, .1]] * 3)
-        rot_tolerance = Tolerance.RotationAxisAngle.default()
+        rot_tolerance = Tolerance.RotationAbsolute.default()
         base_tolerance = pos_tolerance + rot_tolerance
-        constraint_base = Constraints.BasePlacement(ToleratedPose(Transformation.neutral(), base_tolerance))
+        constraint_base = Constraints.BasePlacement(ToleratedPose(WORLD_FRAME, base_tolerance))
         all_goals = Constraints.AllGoalsFulfilled()
 
         # Set up two goals
-        at_goal = Goals.At('At pos z', ToleratedPose(spatial.homogeneous([.1, .1, .5])),
+        at_goal = Goals.At('At pos z',
+                           ToleratedPose(Frame("", Transformation.from_translation([.1, .1, .5]), WORLD_FRAME)),
                            constraints=[constraint_joints])
-        reach_goal = Goals.Reach('Reach z', ToleratedPose(spatial.homogeneous([.1, .1, .5])),
+        reach_goal = Goals.Reach('Reach z',
+                                 ToleratedPose(Frame("", Transformation.from_translation([.1, .1, .5]), WORLD_FRAME)),
                                  constraints=[constraint_joints])
 
         # Combine the goals in a new task (without obstacles for now)
@@ -387,9 +243,9 @@ class TestSolution(unittest.TestCase):
         robot_assembly = get_six_axis_assembly()
         qs = []  # 4 self-collision free robot configurations
         for _ in range(4):
-            q = robot_assembly.robot.random_configuration()
+            q = robot_assembly.robot.random_configuration(self.rng)
             while robot_assembly.robot.has_self_collision(q):
-                q = robot_assembly.robot.random_configuration()
+                q = robot_assembly.robot.random_configuration(self.rng)
             qs.append(q)
         pose_trajectory = Trajectory(pose=np.asarray(tuple(ToleratedPose(robot_assembly.robot.fk(q)) for q in qs)))
         timed_pose_trajectory = Trajectory(pose=pose_trajectory.pose, t=np.asarray((0., .1, .2, .3)))
@@ -407,7 +263,7 @@ class TestSolution(unittest.TestCase):
             # Still ok with additional qs
             valid_sol = SolutionTrajectory(
                 Trajectory(t=np.asarray((0., .1, .15, .2, .3)),
-                           q=np.asarray((*qs[:2], robot_assembly.robot.random_configuration(), *qs[2:])),
+                           q=np.asarray((*qs[:2], robot_assembly.robot.random_configuration(self.rng), *qs[2:])),
                            goal2time={"0": 0.3}),
                 SolutionHeader("tmp"), task, robot_assembly, QDist(), Transformation.neutral())
             self.assertTrue(follow_goal.achieved(valid_sol))
@@ -428,7 +284,7 @@ class TestSolution(unittest.TestCase):
             # Break by changing timing
             valid_sol = SolutionTrajectory(
                 Trajectory(t=np.asarray((0., .1, .15, .23, .4)),
-                           q=np.asarray((*qs[:2], robot_assembly.robot.random_configuration(), *qs[2:])),
+                           q=np.asarray((*qs[:2], robot_assembly.robot.random_configuration(self.rng), *qs[2:])),
                            goal2time={"0": 0.4}),
                 SolutionHeader("tmp"), task, robot_assembly, QDist(), Transformation.neutral())
             if traj.is_timed:
@@ -441,7 +297,7 @@ class TestSolution(unittest.TestCase):
             # Break altering q
             valid_sol = SolutionTrajectory(
                 Trajectory(t=np.asarray((0., .1, .2, .3)),
-                           q=np.asarray((*qs[:2], robot_assembly.robot.random_configuration(), *qs[3:])),
+                           q=np.asarray((*qs[:2], robot_assembly.robot.random_configuration(self.rng), *qs[3:])),
                            goal2time={"0": 0.3}),
                 SolutionHeader("tmp"), task, robot_assembly, QDist(), Transformation.neutral())
             self.assertFalse(follow_goal.achieved(valid_sol))
@@ -452,7 +308,7 @@ class TestSolution(unittest.TestCase):
 
         # Test constraints on whole duration with adding / removing self-collision during execution
         while not robot_assembly.robot.has_self_collision():
-            robot_assembly.robot.configuration = robot_assembly.robot.random_configuration()
+            robot_assembly.robot.update_configuration(robot_assembly.robot.random_configuration(self.rng))
         q_self_collision = robot_assembly.robot.configuration
         follow_goal_with_constraint = Goals.Follow("0", traj, constraints=(Constraints.SelfCollisionFree(),))
         task = Task.Task(TaskHeader("tmp", timeStepSize=0.1), goals=(follow_goal_with_constraint,))
@@ -465,9 +321,9 @@ class TestSolution(unittest.TestCase):
             valid_sol, valid_sol.t_goals["0"], len(valid_sol.trajectory)))
 
         # Adding a random, non-colliding should be fine without time
-        q_rand_coll_free = robot_assembly.robot.random_configuration()
+        q_rand_coll_free = robot_assembly.robot.random_configuration(self.rng)
         while robot_assembly.robot.has_self_collision(q_rand_coll_free):
-            q_rand_coll_free = robot_assembly.robot.random_configuration()
+            q_rand_coll_free = robot_assembly.robot.random_configuration(self.rng)
         validish_sol = SolutionTrajectory(
             Trajectory(t=np.asarray((0., .1, .2, .3, .4)),
                        q=np.asarray((*qs[:2], q_rand_coll_free, *qs[2:])), goal2time={"0": 0.4}),
@@ -490,6 +346,55 @@ class TestSolution(unittest.TestCase):
         self.assertFalse(follow_goal_with_constraint._valid(
             invalid_sol, invalid_sol.t_goals["0"], len(invalid_sol.trajectory)))
 
+    def test_follow_force_torque_constraint(self):
+        assembly = get_six_axis_assembly()
+        robot = assembly.robot
+        robot.model.effortLimit = 10 * np.ones(robot.dof)  # high torque limits
+        len_desired = 15
+        len_additional = 5
+        dt = .1
+        q = []
+        while len(q) < len_desired + len_additional:
+            new = q[-1] + self.rng.random((robot.dof,)) * 0.01 if len(q) > 0 \
+                else robot.random_configuration(self.rng)
+            if not robot.has_self_collision(new) \
+                    and robot.tau_in_torque_limits(robot.static_torque(new)) \
+                    and robot.q_in_joint_limits(new):
+                q.append(new)
+
+        desired = Trajectory(pose=np.asarray(tuple(ToleratedPose(robot.fk(qi)) for qi in q[:len_desired])), t=dt)
+        follow_goal_with_constraint = Goals.Follow("0", desired,
+                                                   constraints=(Constraints.JointLimits(parts=('q',)),))
+        task = Task.Task(TaskHeader("tmp", timeStepSize=dt),
+                         goals=(follow_goal_with_constraint,),
+                         constraints=(Constraints.JointLimits(parts=('q', 'tau')),
+                                      Constraints.AllGoalsFulfilled(),
+                                      Constraints.CollisionFree()
+                                      )
+                         )
+
+        trajectory = Trajectory(q=np.asarray(q), t=dt, goal2time={"0": (len_desired - 1) * dt})
+        solution = SolutionTrajectory(trajectory, {'taskID': task.id}, task, assembly, QDist())
+
+        # Thus far, the solution should be valid (assuming the robot is not moving too fast yet)
+        self.assertTrue(solution.valid)
+
+        # Add high external force
+        follow_goal_with_constraint._external_forces = self.rng.random((3,)) * 1000
+        solution = SolutionTrajectory(trajectory, {'taskID': task.id}, task, assembly, QDist())
+        self.assertFalse(solution.valid)
+        follow_goal_with_constraint._external_forces = np.zeros((3,))  # reset
+
+        # Add high external torques
+        follow_goal_with_constraint._external_torques = self.rng.random((3,)) * 1000
+        solution = SolutionTrajectory(trajectory, {'taskID': task.id}, task, assembly, QDist())
+        self.assertFalse(solution.valid)
+
+        # Remove the torque constraint
+        task.constraints = (Constraints.AllGoalsFulfilled(), Constraints.CollisionFree())
+        solution = SolutionTrajectory(trajectory, {'taskID': task.id}, task, assembly, QDist())
+        self.assertTrue(solution.valid)
+
     def test_solution_derived_properties(self, num_test=10, epsilon=1e-12):
         assembly = get_six_axis_assembly()  # Use IMPROV with significant joint inertia and friction
         for _ in range(num_test):
@@ -526,9 +431,9 @@ class TestSolution(unittest.TestCase):
             robot.update_configuration(q1)
             viz = task.visualize(viz, robots=robot, center_view=False)  # Should have view centered on origin
             clear_visualizer(viz)
-            viz = task.visualize(viz, robots=robot, center_view=True)  # Should have view including more of the robot
+            viz = task.visualize(viz, robots=robot, center_view=True)  # Should have a view including more of the robot
             clear_visualizer(viz)
-            viz = task.visualize(viz, center_view=True)  # Should have view including base constraint, e.g. PTP_1
+            viz = task.visualize(viz, center_view=True)  # Should have a view including base constraint, e.g. PTP_1
             clear_visualizer(viz)
             robot.update_configuration(q2)
             self.assertIs(robot.data, viz.data)
