@@ -9,6 +9,7 @@ import numpy.testing as np_test
 from timor import Transformation
 from timor.utilities import logging
 from timor.utilities.dtypes import Lazy
+from timor.utilities.frames import Frame, FrameTree, WORLD_FRAME
 from timor.utilities.tolerated_pose import ToleratedPose
 from timor.utilities.trajectory import Trajectory
 import timor.utilities.errors as err
@@ -84,6 +85,24 @@ class TestTrajectoryClasses(unittest.TestCase):
         self.assertEqual(traj_with_lazy_dq, traj_with_lazy_dq + empty)
         self.assertEqual(traj_with_lazy_dq, empty + traj_with_lazy_dq)
 
+    def test_stationary(self):
+        """Test stationary trajectory."""
+        q = np.random.random((4,))
+        t_length = 1.
+        dt = 0.07
+        traj = Trajectory.stationary(t_length, q=q)
+        self.assertEqual(len(traj), 2)
+        traj_tight = Trajectory.stationary(t_length, q=q, sample_time=dt)
+        self.assertGreater(len(traj_tight), len(traj))
+        for t in (traj, traj_tight):
+            self.assertEqual(t.t[0], 0.)
+            self.assertGreaterEqual(t.t[-1], t_length)
+            self.assertGreaterEqual(t_length + dt, t.t[-1])
+            for idx in range(len(t)):
+                np_test.assert_array_equal(t.q[idx], q)
+                np_test.assert_array_equal(t.dq[idx], np.zeros_like(q))
+                np_test.assert_array_equal(t.ddq[idx], np.zeros_like(q))
+
     def test_untimed(self):
         """Based on issue #23."""
         arr = np.random.rand(100, 1)
@@ -146,16 +165,16 @@ class TestTrajectoryClasses(unittest.TestCase):
 
     def test_PoseTrajectory(self):
         example_pose_trajectory = Trajectory(pose=np.asarray([
-            ToleratedPose(Transformation.neutral()),
-            ToleratedPose(Transformation.from_translation((0, 0, 1))),
-            ToleratedPose(Transformation.from_translation((0, 1, 1)))]))
+            ToleratedPose(Frame("", Transformation.neutral(), WORLD_FRAME)),
+            ToleratedPose(Frame("", Transformation.from_translation((0, 0, 1)), WORLD_FRAME)),
+            ToleratedPose(Frame("", Transformation.from_translation((0, 1, 1)), WORLD_FRAME))]))
         example_timed_pose_trajectory = Trajectory(t=np.asarray((0., 0.1, 0.2)),
                                                    pose=example_pose_trajectory.pose,
                                                    _allowed_deviation_in_time=0.01)
 
         self.assertEqual(3, len(example_pose_trajectory))
         self.assertTrue(example_pose_trajectory[2].pose[0].valid(Transformation.from_translation((0, 0.9999, 1))))
-        t_new = Trajectory.from_json_data(example_pose_trajectory.to_json_data())
+        t_new = Trajectory.from_json_data(example_pose_trajectory.to_json_data(), frames=FrameTree.empty())
         self.assertEqual(example_pose_trajectory, t_new)
         with self.assertRaises(ValueError):  # len(t) != len(pose)
             _ = Trajectory(t=np.asarray((0., 0.1, 0.2, 0.3)),
@@ -166,12 +185,14 @@ class TestTrajectoryClasses(unittest.TestCase):
         with self.assertRaises(ValueError):
             example_pose_trajectory.get_at_time(0.1)
         self.assertTrue(example_timed_pose_trajectory.get_at_time(0.1).pose[0].
-                        valid(example_pose_trajectory.pose[1].nominal))
+                        valid(example_pose_trajectory.pose[1].nominal.in_world_coordinates()))
         # Check access via index instead of time
-        self.assertTrue(example_timed_pose_trajectory[1].pose[0].valid(example_pose_trajectory.pose[1].nominal))
+        self.assertTrue(
+            example_timed_pose_trajectory[1].pose[0].valid(
+                example_pose_trajectory.pose[1].nominal.in_world_coordinates()))
         self.assertFalse(example_timed_pose_trajectory.get_at_time(0.1).pose[0].
-                         valid(example_pose_trajectory.pose[2].nominal))
-        t_timed_new = Trajectory.from_json_data(example_timed_pose_trajectory.to_json_data())
+                         valid(example_pose_trajectory.pose[2].nominal.in_world_coordinates()))
+        t_timed_new = Trajectory.from_json_data(example_timed_pose_trajectory.to_json_data(), frames=FrameTree.empty())
         self.assertEqual(example_timed_pose_trajectory, t_timed_new)
         self.assertNotEqual(example_pose_trajectory, example_timed_pose_trajectory)
 
